@@ -19,13 +19,15 @@ from cliff import lister
 from osc_lib.command import command
 from osc_lib import exceptions
 from osc_lib import utils
+from osc_lib.utils import tags as _tag
 from oslo_utils import uuidutils
 
 from octaviaclient.osc.v2 import constants as const
 from octaviaclient.osc.v2 import utils as v2_utils
 from octaviaclient.osc.v2 import validate
 
-PROTOCOL_CHOICES = ['TCP', 'HTTP', 'HTTPS', 'TERMINATED_HTTPS', 'UDP']
+PROTOCOL_CHOICES = ['TCP', 'HTTP', 'HTTPS', 'TERMINATED_HTTPS', 'UDP', 'SCTP',
+                    'PROMETHEUS']
 CLIENT_AUTH_CHOICES = ['NONE', 'OPTIONAL', 'MANDATORY']
 
 
@@ -33,7 +35,7 @@ class CreateListener(command.ShowOne):
     """Create a listener"""
 
     def get_parser(self, prog_name):
-        parser = super(CreateListener, self).get_parser(prog_name)
+        parser = super().get_parser(prog_name)
 
         parser.add_argument(
             'loadbalancer',
@@ -173,7 +175,7 @@ class CreateListener(command.ShowOne):
         parser.add_argument(
             '--wait',
             action='store_true',
-            help='Wait for action to complete',
+            help='Wait for action to complete.',
         )
         parser.add_argument(
             '--tls-ciphers',
@@ -181,6 +183,27 @@ class CreateListener(command.ShowOne):
             help="Set the TLS ciphers to be used "
                  "by the listener in OpenSSL format."
         )
+        parser.add_argument(
+            '--tls-version',
+            dest='tls_versions',
+            metavar='<tls_versions>',
+            nargs='?',
+            action='append',
+            help="Set the TLS protocol version to be used "
+                 "by the listener (can be set multiple times)."
+        )
+        parser.add_argument(
+            '--alpn-protocol',
+            dest='alpn_protocols',
+            metavar='<alpn_protocols>',
+            nargs='?',
+            action='append',
+            help="Set the ALPN protocol to be used "
+                 "by the listener (can be set multiple times)."
+        )
+
+        _tag.add_tag_option_to_parser_for_create(
+            parser, 'listener')
 
         return parser
 
@@ -211,7 +234,8 @@ class CreateListener(command.ShowOne):
                       'pools': v2_utils.format_list,
                       'l7policies': v2_utils.format_list,
                       'insert_headers': v2_utils.format_hash,
-                      'allowed_cidrs': v2_utils.format_list_flat}
+                      'allowed_cidrs': v2_utils.format_list_flat,
+                      'tags': v2_utils.format_list_flat}
 
         return (rows,
                 (utils.get_dict_properties(data['listener'],
@@ -223,17 +247,17 @@ class DeleteListener(command.Command):
     """Delete a listener"""
 
     def get_parser(self, prog_name):
-        parser = super(DeleteListener, self).get_parser(prog_name)
+        parser = super().get_parser(prog_name)
 
         parser.add_argument(
             'listener',
             metavar="<listener>",
-            help="Listener to delete (name or ID)"
+            help="Listener to delete (name or ID)."
         )
         parser.add_argument(
             '--wait',
             action='store_true',
-            help='Wait for action to complete',
+            help='Wait for action to complete.',
         )
 
         return parser
@@ -258,7 +282,7 @@ class ListListener(lister.Lister):
     """List listeners"""
 
     def get_parser(self, prog_name):
-        parser = super(ListListener, self).get_parser(prog_name)
+        parser = super().get_parser(prog_name)
 
         parser.add_argument(
             '--name',
@@ -290,6 +314,9 @@ class ListListener(lister.Lister):
             metavar='<project>',
             help="List listeners by project ID."
         )
+
+        _tag.add_tag_filtering_option_to_parser(parser, 'listener')
+
         return parser
 
     def take_action(self, parsed_args):
@@ -307,12 +334,12 @@ class ShowListener(command.ShowOne):
     """Show the details of a single listener"""
 
     def get_parser(self, prog_name):
-        parser = super(ShowListener, self).get_parser(prog_name)
+        parser = super().get_parser(prog_name)
 
         parser.add_argument(
             'listener',
             metavar='<listener>',
-            help='Name or UUID of the listener'
+            help='Name or UUID of the listener.'
         )
 
         return parser
@@ -339,7 +366,8 @@ class ShowListener(command.ShowOne):
                       'pools': v2_utils.format_list,
                       'l7policies': v2_utils.format_list,
                       'insert_headers': v2_utils.format_hash,
-                      'allowed_cidrs': v2_utils.format_list_flat}
+                      'allowed_cidrs': v2_utils.format_list_flat,
+                      'tags': v2_utils.format_list_flat}
 
         return rows, utils.get_dict_properties(data, rows,
                                                formatters=formatters)
@@ -349,7 +377,7 @@ class SetListener(command.Command):
     """Update a listener"""
 
     def get_parser(self, prog_name):
-        parser = super(SetListener, self).get_parser(prog_name)
+        parser = super().get_parser(prog_name)
 
         parser.add_argument(
             'listener',
@@ -383,7 +411,7 @@ class SetListener(command.Command):
             '--default-tls-container-ref',
             metavar='<container-ref>',
             help="The URI to the key manager service secrets container "
-                 "containing the certificate and key for TERMINATED_TLS"
+                 "containing the certificate and key for TERMINATED_TLS "
                  "listeners."
         )
         parser.add_argument(
@@ -474,7 +502,7 @@ class SetListener(command.Command):
         parser.add_argument(
             '--wait',
             action='store_true',
-            help='Wait for action to complete',
+            help='Wait for action to complete.',
         )
         parser.add_argument(
             '--tls-ciphers',
@@ -482,6 +510,26 @@ class SetListener(command.Command):
             help="Set the TLS ciphers to be used "
                  "by the listener in OpenSSL format."
         )
+        parser.add_argument(
+            '--tls-version',
+            dest='tls_versions',
+            metavar='<tls_versions>',
+            nargs='?',
+            action='append',
+            help="Set the TLS protocol version to be used "
+                 "by the listener (can be set multiple times)."
+        )
+        parser.add_argument(
+            '--alpn-protocol',
+            dest='alpn_protocols',
+            metavar='<alpn_protocols>',
+            nargs='?',
+            action='append',
+            help="Set the ALPN protocol to be used "
+                 "by the listener (can be set multiple times)."
+        )
+
+        _tag.add_tag_option_to_parser_for_set(parser, 'listener')
 
         return parser
 
@@ -491,6 +539,10 @@ class SetListener(command.Command):
 
         listener_id = attrs.pop('listener_id')
 
+        v2_utils.set_tags_for_set(
+            self.app.client_manager.load_balancer.listener_show,
+            listener_id, attrs, clear_tags=parsed_args.no_tag)
+
         body = {'listener': attrs}
 
         self.app.client_manager.load_balancer.listener_set(
@@ -507,7 +559,7 @@ class UnsetListener(command.Command):
     """Clear listener settings"""
 
     def get_parser(self, prog_name):
-        parser = super(UnsetListener, self).get_parser(prog_name)
+        parser = super().get_parser(prog_name)
 
         parser.add_argument(
             'listener',
@@ -592,22 +644,44 @@ class UnsetListener(command.Command):
             action='store_true',
             help="Clear all allowed CIDRs from the listener."
         )
+        parser.add_argument(
+            '--tls-versions',
+            action='store_true',
+            help='Clear all TLS versions from the listener.',
+        )
+        parser.add_argument(
+            '--tls-ciphers',
+            action='store_true',
+            help='Clear all TLS ciphers from the listener.',
+        )
         parser.add_argument(
             '--wait',
             action='store_true',
-            help='Wait for action to complete',
+            help='Wait for action to complete.',
         )
+        parser.add_argument(
+            '--alpn-protocols',
+            action='store_true',
+            help="Clear all ALPN protocols from the listener."
+        )
+
+        _tag.add_tag_option_to_parser_for_unset(parser, 'listener')
+
         return parser
 
     def take_action(self, parsed_args):
         unset_args = v2_utils.get_unsets(parsed_args)
-        if not unset_args:
+        if not unset_args and not parsed_args.all_tag:
             return
 
         listener_id = v2_utils.get_resource_id(
             self.app.client_manager.load_balancer.listener_list,
             'listeners', parsed_args.listener)
 
+        v2_utils.set_tags_for_unset(
+            self.app.client_manager.load_balancer.listener_show,
+            listener_id, unset_args, clear_tags=parsed_args.all_tag)
+
         body = {'listener': unset_args}
 
         self.app.client_manager.load_balancer.listener_set(
@@ -624,12 +698,12 @@ class ShowListenerStats(command.ShowOne):
     """Shows the current statistics for a listener."""
 
     def get_parser(self, prog_name):
-        parser = super(ShowListenerStats, self).get_parser(prog_name)
+        parser = super().get_parser(prog_name)
 
         parser.add_argument(
             'listener',
             metavar='<listener>',
-            help='Name or UUID of the listener'
+            help='Name or UUID of the listener.'
         )
 
         return parser
